'''
classes for gcskewer
'''
class Frame:
    '''
    frame for the gc plots
    '''
    cummulative_gc = 0
    cummulative_at = 0
    header = [
        'record',
        'sequence',
        'start',
        'end',
        'midpoint',
        'g_count',
        'c_count',
        't_count',
        'a_count',
        'gc_content',
        'at_content',
        'gc_skew',
        'at_skew',
        'cummulative_gc',
        'cummulative_at'
    ] #this can be automated more safely

    @classmethod
    def reset(cls):
        '''
        reset the class variables
        useful when moving to a new record
        '''
        cls.cummulative_gc = 0
        cls.cummulative_at = 0

    def __init__(
        self, record: str, sequence: str, start: int, end: int
        ):
        '''
        initialisation function
        '''
        self.record = record
        self.sequence = sequence
        self.start = start
        self.end = end
        self.midpoint = start + (start + end) / 2
        self.g_count = self.count_nt('G')
        self.c_count = self.count_nt('C')
        self.t_count = self.count_nt('T')
        self.a_count =  self.count_nt('A')
        self.gc_content = (self.g_count + self.c_count) / (len(self.sequence))
        self.at_content = 1 - self.gc_content
        self.gc_skew = self.calculate_skew(self.g_count, self.c_count)
        self.at_skew = self.calculate_skew(self.a_count, self.t_count)

        type(self).cummulative_gc += self.gc_skew
        type(self).cummulative_at += self.at_skew
        self.cummulative_gc = type(self).cummulative_gc
        self.cummulative_at = type(self).cummulative_at

    def count_nt(self, nt: str) -> int:
        '''
        count the occurance of a letter in the string
        '''
        nt = nt.upper()
        return self.sequence.upper().count(nt)

    def calculate_skew(self, x: int, y: int) -> float:
        '''
        calculate gc skew
            arguments
        '''
        skew = (x - y)/(x + y)
        return skew

    def get_line(self):
        '''
        return a list of class variables
        '''
        return list(self.__dict__.values())



'''
parser for gcskewer
    functions:
        get_parser()
        get_args()
'''

import argparse
from argparse import RawTextHelpFormatter

def get_parser():
    '''
    gets argument parser
        arguments:
            None
        returns:
            argument_parser
    '''
    parser = argparse.ArgumentParser(
        prog='gcskewer',
        description=
            '''
            Calculates and plots gc-skew from DNA sequences.
                EXAMPLE USAGE 1:
                    gcskewer -f my_fasta.fna -p 
                EXAMPLE USAGE 2:
                    gcskewer -g my_genbank.gbk -ws 1000 -ss 100 -c
            ''',
        epilog='Written by Dr. T. J. Booth, 2023',
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument(
        '-g',
        '--genbank',
        type=str,
        default=None,
        help = (
            'path to input genbank file'
        )
    )
    parser.add_argument(
        '-f',
        '--fasta',
        type=str,
        default=None,
        help = (
            'path to input fasta nucleotide file'
        )
    )
    parser.add_argument(
        '-ws',
        '--window-size',
        type=int,
        default=None,
        help = (
            'window size for gc-skew plot'
        )
    )
    parser.add_argument(
        '-ss',
        '--step-size',
        type=int,
        default=None,
        help = (
            'step size for gc-skew plot'
        )
    )
    parser.add_argument(
        '-c',
        '--csv',
        action="store_true",
        help = (
            'write the output to a .csv file'
        )
    )
    parser.add_argument(
        '-s',
        '--svg',
        action="store_true",
        help = (
            'write the output to an .svg file'
        )
    )
    parser.add_argument(
        '-p',
        '--plot',
        action="store_true",
        help = (
            'write the output to an .html file'
        )
    )
    return parser

def get_args():
    '''
    gets the arguments from the parser
        arguments:
            None
        returns:
            args: arguments object
    '''
    parser = get_parser()
    args = parser.parse_args()
    return args



'''
errors for gcskewer
'''
class GCSkewerError(Exception):
    '''
    General error for gcskewer
    '''

class InputError(GCSkewerError):
    '''
    Error raised for incorrectly provided inputs
    '''

class OutputError(GCSkewerError):
    '''
    Error raised when no output is provided
    '''



'''
main script for gcskewer
'''
from typing import Dict, List

from gcskewer import checks, output, parser
from gcskewer.classes.dataclasses import Frame

def get_dict_for_plotting(frames: List[Frame]) -> Dict:
    '''
    gets data as a dictionary from members of the Frame class
        arguments:
            frames: a list of Fram objects
        returns:
            plotting_dict: frame data formmatted as a Dict
    '''
    plotting_dict = {
        'midpoint' :  [frame.midpoint for frame in frames],  
        'gc_skew' : [frame.gc_skew for frame in frames],
        'at_skew' : [frame.at_skew for frame in frames],
        'cummulative_gc_skew' : [frame.cummulative_gc for frame in frames],
        'cummulative_at_skew' : [frame.cummulative_at for frame in frames],
        'gc_content' : [frame.gc_content for frame in frames],
        'at_content' : [frame.at_content for frame in frames],
    }
    return plotting_dict

def get_frames(
    record_name: str, record_sequence: str, window_size: int, step_size: int
    ) -> List[Frame]:
    '''
    get gc skew and store in a dataframe
        record_name: name of the dna sequence
        record_sequence: dna sequence
        window_size: size of the analysis window
        step_size: step size to move the window
    returns:
        gc_dataframe: gc data dependent on midpoint for plotting
    '''
    skew_data = []
    sequence_length = len(record_sequence) - window_size
    output.print_to_system('Calculating GC-skew for ' + record_name + '.')
    #calculate GC skew for each subsequence using the sliding window
    for start_point in range(0, sequence_length, step_size):
        end_point = start_point + window_size
        sub_sequence = record_sequence[start_point:end_point]
        frame = Frame(record_name, sub_sequence, start_point, end_point)
        skew_data.append(frame)
    return skew_data

def main():
    '''
    main routine for gcskewer
        arguments: 
            None
        returns: 
            None
    '''
    output.print_to_system('Running gcskewer version 1.1.0')
    args = parser.get_args()
    checks.check_output(args)
    filename, _format = checks.check_input(args)
    records = output.read_file(filename, _format)
    assert records, (
        'No input sequences!'
        'Check your file and that you have used the correct parameter for the filetype (-g/-f)'
    )
    window_size, step_size = checks.check_window_and_step(
        args.window_size, args.step_size, records.values()
        )
    for record_name, record_sequence in records.items():
        frames = get_frames(record_name, record_sequence, window_size, step_size)
        name = record_name
        # technically frame_dict doesnt need to be calculated for .csv
        frame_dict = get_dict_for_plotting(frames)
        if args.csv:
            output.write_gcframes_to_csv(frames, name)
        if args.plot:
            output.plot_html(frame_dict, name)
        if args.svg:
            output.plot_svg(frame_dict, name)
        Frame.reset() #vital!
    output.print_to_system('gcskewer has finished!')

if __name__ == "__main__":
    main()



'''
checks inputs for gcskewer
    functions:
    !!!!!!!!!
'''
from typing import List
from gcskewer.output import print_to_system
from gcskewer.classes.errors import InputError, OutputError

def check_output(args):
    '''
    check an output exists
    '''
    output_args = [args.csv, args.svg, args.plot]
    if all(arg is False for arg in output_args):
        raise OutputError(
            'No output flag was set! Please choose at least 1 output.'
            )

def check_input(args):
    '''
    Checks what input the user provided, and
    errors if neither or both is set, or
    returns file type for reading
        arguments:
            args: args from argsparse
        returns:
            input path, file extension
    '''
    if args.genbank is None and args.fasta is None:
        raise InputError('No input file!')
    if args.fasta and args.genbank:
        raise InputError('Both genbank and fasta input provided - pick one!')
    if args.genbank is None:
        return args.fasta, 'fasta'
    if args.fasta is None:
        return args.genbank, 'genbank'
    raise InputError(
        'Something unexpected happened when processing the input. Please report the issue!'
        )

def get_window_size(sequences: List):
    '''
    return window size of 1% of the smalles sequences
        arguments:
            record_sequences: list of dna sequences
        returns:
            window_size: 1% of the smallest sequence
    '''
    sequence_lengths = [len(seq) for seq in sequences]
    window_size = min(sequence_lengths)//100
    return window_size

def get_step_size(window_size: int):
    '''
    returns 1/10 of the window size
        arguments:
            window_size: window size for analysis
        returns:
            step_size: step size is 1/10 of the window size
    '''
    step_size = window_size // 10
    return step_size

def check_window_and_step(window_size, step_size, sequences):
    '''
    checks if window and step sizes are set and, 
    if not, automatically sets them proportionate to the size of the smallest input sequence
        arguments:
            window_size: window size from argparse
            step_size: step size from argparse
            sequences: a list of input dna sequences
    '''
    if window_size is None:
        window_size = get_window_size(sequences)
        print_to_system(
            'Warning: No window size provided. Automatically set to: ' + str(window_size)
            )
    if step_size is None:
        step_size = get_step_size(window_size)
        print_to_system('Warning: No step size provided. Automatically set to: ' + str(step_size))
    return window_size, step_size



'''
reading and writing for gcskewer
    functions:
        read_file(filename:str, _format: str)-> List[str], List[str]
        write_csv(df: DataFrame, name:str) -> None
        plot_data(df: DataFrame, name: str) -> None
        plot_svg(df: DataFrame, name: str) -> None
        print_to_system(string_to_print: str) -> None
'''
import csv
from datetime import datetime
from typing import List, Dict

from Bio import SeqIO
from plotly.subplots import make_subplots
import matplotlib.pyplot as plt
import plotly.graph_objects as go

from gcskewer.classes.dataclasses import Frame

def read_file(filename:str, _format: str) -> (List[str], List[str]):
    '''
    read file with biopython and reurn record names and sequences as two lists
        arguments:
            filename: path to the file to be read
            _format: string specifying 'fasta' or 'genbank' for BioPython
        returns:
            records: a dictionary of {record.id : record.seq}
    '''
    print_to_system('Reading fasta: ' + filename)
    records = {}
    for seq_record in SeqIO.parse(filename, _format):
        records.update({seq_record.id : seq_record.seq})
    return records

def write_gcframes_to_csv(frames: List[Frame], name:str) -> None:
    '''
    writes the df to a .csv
        arguments: 
            frames: a list of Frame objects
            name: the name of the record the dataframe was generated from
        returns:
            None
    '''
    lines = [Frame.header]
    lines.extend([frame.get_line() for frame in frames])
    filename = name + '.csv'
    print_to_system('Writing to ' + filename)
    with open(filename, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerows(lines)

def plot_html(frames: Dict, name: str) -> None:
    '''
    plot the data to an .html file with plotly
         arguments:
            frames: the dictionary of frame data
            name: the name of the record from which the df originated
        returns:
            None
    '''
    fig = make_subplots(specs=[[{"secondary_y": True}]])
    ### SKEW ###
    fig.add_trace(
        go.Scatter(x=frames['midpoint'], y=frames['gc_skew'], name='GC Skew'), secondary_y=False
        )
    fig.add_trace(
        go.Scatter(x=frames['midpoint'], y=frames['at_skew'], name='AT Skew'), secondary_y=False
        )
    fig.add_trace(
        go.Scatter(
            x=frames['midpoint'], y=frames['cummulative_gc_skew'], name='Cummulative GC Skew'
            ), secondary_y=True
        )
    fig.add_trace(
        go.Scatter(
            x=frames['midpoint'], y=frames['cummulative_at_skew'], name='Cummulative AT Skew'
            ), secondary_y=True
        )
    ### CONTENT ###
    fig.add_trace(
        go.Scatter(
            x=frames['midpoint'], y=frames['gc_content'], name='GC Content'
            ), secondary_y=False
        )
    fig.add_trace(
        go.Scatter(
            x=frames['midpoint'], y=frames['at_content'], name='AT Content'
            ), secondary_y=False
        )
    file_name = name + ".html"
    fig.write_html(file_name)
    print_to_system('Plot saved to ' + file_name)

def plot_svg(frames: Frame, name: str) -> None:
    '''
    plot svg from dataframe
        arguments:
            frames: the dictionary of frame data
            name: the name of the record from which the df originated
        returns:
            None
    '''
    ### AXIS 1: SKEW AND CONTENT ###
    _, ax1 = plt.subplots()
    ax1.plot(frames['midpoint'], frames['gc_skew'], label="GC skew", color='#56b4e9')
    ax1.plot(frames['midpoint'], frames['at_skew'], label="AT skew", color='#e69f00')
    ax1.plot(frames['midpoint'], frames['gc_content'], label="GC content", color='#0072B2')
    ax1.plot(frames['midpoint'], frames['at_content'], label="AT content", color='#f0e442')
    ax1.set_xlabel('Position')
    ax1.set_ylabel('Skew/Content')
    ax1.set_title('GC skew of ' + name)
    ### AXIS2: CUMMULATIVE SKEW ###S
    ax2 = plt.twinx()
    ax2.plot(
        frames['midpoint'], frames['cummulative_gc_skew'],
        label="Cumulative GC skew",color='#009e73'
        )
    ax2.plot(
        frames['midpoint'], frames['cummulative_at_skew'],
        label="Cumulative AT skew", color='#cc79a7'
        )
    ax2.set_ylabel('Cummulative Skew')
    #legend
    lines, labels = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    plt.legend(lines + lines2, labels + labels2)
    plt.savefig(name + '.svg')

def print_to_system(string_to_print: str) -> None:
    '''
    prints a line to the terminal with date and time
        arguments:
            string_to_print: the text to be printed
        returns:
            None
    '''
    now = datetime.now()
    current_time = now.strftime("[%H:%M:%S]: ")
    print(current_time + string_to_print)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    description = fh.read()

setup(
    name="gcskewer",
    version="1.1.0",
    author="Thomas J. Booth",
    author_email="thoboo@biosustain.dtu.dk",
    packages=find_packages(),
    description="A python for plotting GC-skew from DNA sequences.",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/DrBoothTJ/skewer",
    license='GNU General Public License v3.0',
    python_requires='>=3.7',
    install_requires=['plotly','Bio'],
    entry_points={'console_scripts': ["gcskewer=gcskewer.main:main"]}
)
